# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCacheBySymbolById, ArrayCacheByTimestamp
from ccxt.base.types import Balances, Int, Order, OrderBook, Str, Strings, Ticker, Tickers, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import NotSupported


class onetrading(ccxt.async_support.onetrading):

    def describe(self):
        return self.deep_extend(super(onetrading, self).describe(), {
            'has': {
                'ws': True,
                'watchBalance': True,
                'watchTicker': True,
                'watchTickers': True,
                'watchTrades': False,
                'watchMyTrades': True,
                'watchOrders': True,
                'watchOrderBook': True,
                'watchOHLCV': True,
            },
            'urls': {
                'api': {
                    'ws': 'wss://streams.onetrading.com/',
                },
            },
            'options': {
                'bp_remaining_quota': 200,
                'timeframes': {
                    '1m': {
                        'unit': 'MINUTES',
                        'period': 1,
                    },
                    '5m': {
                        'unit': 'MINUTES',
                        'period': 5,
                    },
                    '15m': {
                        'unit': 'MINUTES',
                        'period': 15,
                    },
                    '30m': {
                        'unit': 'MINUTES',
                        'period': 30,
                    },
                    '1h': {
                        'unit': 'HOURS',
                        'period': 1,
                    },
                    '4h': {
                        'unit': 'HOURS',
                        'period': 4,
                    },
                    '1d': {
                        'unit': 'DAYS',
                        'period': 1,
                    },
                    '1w': {
                        'unit': 'WEEKS',
                        'period': 1,
                    },
                    '1M': {
                        'unit': 'MONTHS',
                        'period': 1,
                    },
                },
            },
            'streaming': {
            },
            'exceptions': {
            },
        })

    async def watch_balance(self, params={}) -> Balances:
        """
        :see: https://developers.bitpanda.com/exchange/#account-history-channel
        watch balance and get the amount of funds available for trading or funds locked in orders
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        await self.authenticate(params)
        url = self.urls['api']['ws']
        messageHash = 'balance'
        subscribeHash = 'ACCOUNT_HISTORY'
        bpRemainingQuota = self.safe_integer(self.options, 'bp_remaining_quota', 200)
        subscribe: dict = {
            'type': 'SUBSCRIBE',
            'bp_remaining_quota': bpRemainingQuota,
            'channels': [
                {
                    'name': 'ACCOUNT_HISTORY',
                },
            ],
        }
        request = self.deep_extend(subscribe, params)
        return await self.watch(url, messageHash, request, subscribeHash, request)

    def handle_balance_snapshot(self, client, message):
        #
        # snapshot
        #     {
        #         "account_id": "b355abb8-aaae-4fae-903c-c60ff74723c6",
        #         "type": "BALANCES_SNAPSHOT",
        #         "channel_name": "ACCOUNT_HISTORY",
        #         "time": "2019-04-01T13:39:17.155Z",
        #         "balances": [{
        #                 "account_id": "b355abb8-aaae-4fae-903c-c60ff74723c6",
        #                 "currency_code": "BTC",
        #                 "change": "0.5",
        #                 "available": "10.0",
        #                 "locked": "1.1234567",
        #                 "sequence": 1,
        #                 "time": "2019-04-01T13:39:17.155Z"
        #             },
        #             {
        #                 "account_id": "b355abb8-aaae-4fae-903c-c60ff74723c6",
        #                 "currency_code": "ETH",
        #                 "change": "0.5",
        #                 "available": "10.0",
        #                 "locked": "1.1234567",
        #                 "sequence": 2,
        #                 "time": "2019-04-01T13:39:17.155Z"
        #             }
        #         ]
        #     }
        #
        self.balance = self.parse_balance(message)
        messageHash = 'balance'
        client.resolve(self.balance, messageHash)

    async def watch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        :see: https://developers.bitpanda.com/exchange/#market-ticker-channel
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        subscriptionHash = 'MARKET_TICKER'
        messageHash = 'ticker.' + symbol
        request: dict = {
            'type': 'SUBSCRIBE',
            'channels': [
                {
                    'name': 'MARKET_TICKER',
                    'price_points_mode': 'INLINE',
                },
            ],
        }
        return await self.watch_many(messageHash, request, subscriptionHash, [symbol], params)

    async def watch_tickers(self, symbols: Strings = None, params={}) -> Tickers:
        """
        :see: https://developers.bitpanda.com/exchange/#market-ticker-channel
        watches price tickers, a statistical calculation with the information for all markets or those specified.
        :param str symbols: unified symbols of the markets to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: an array of `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        symbols = self.market_symbols(symbols)
        if symbols is None:
            symbols = []
        subscriptionHash = 'MARKET_TICKER'
        messageHash = 'tickers'
        request: dict = {
            'type': 'SUBSCRIBE',
            'channels': [
                {
                    'name': 'MARKET_TICKER',
                    'price_points_mode': 'INLINE',
                },
            ],
        }
        tickers = await self.watch_many(messageHash, request, subscriptionHash, symbols, params)
        return self.filter_by_array(tickers, 'symbol', symbols)

    def handle_ticker(self, client: Client, message):
        #
        #     {
        #         "ticker_updates": [{
        #             "instrument": "ETH_BTC",
        #             "last_price": "0.053752",
        #             "price_change": "0.000623",
        #             "price_change_percentage": "1.17",
        #             "high": "0.055",
        #             "low": "0.052662",
        #             "volume": "6.3821593247"
        #         }],
        #         "channel_name": "MARKET_TICKER",
        #         "type": "MARKET_TICKER_UPDATES",
        #         "time": "2022-06-23T16:41:00.004162Z"
        #     }
        #
        tickers = self.safe_value(message, 'ticker_updates', [])
        datetime = self.safe_string(message, 'time')
        for i in range(0, len(tickers)):
            ticker = tickers[i]
            marketId = self.safe_string(ticker, 'instrument')
            symbol = self.safe_symbol(marketId)
            self.tickers[symbol] = self.parse_ws_ticker(ticker)
            timestamp = self.parse8601(datetime)
            self.tickers[symbol]['timestamp'] = timestamp
            self.tickers[symbol]['datetime'] = self.iso8601(timestamp)
            client.resolve(self.tickers[symbol], 'ticker.' + symbol)
        client.resolve(self.tickers, 'tickers')

    def parse_ws_ticker(self, ticker, market=None):
        #
        #     {
        #         "instrument": "ETH_BTC",
        #         "last_price": "0.053752",
        #         "price_change": "-0.000623",
        #         "price_change_percentage": "-1.17",
        #         "high": "0.055",
        #         "low": "0.052662",
        #         "volume": "6.3821593247"
        #     }
        #
        marketId = self.safe_string(ticker, 'instrument')
        return self.safe_ticker({
            'symbol': self.safe_symbol(marketId, market),
            'timestamp': None,
            'datetime': None,
            'high': self.safe_string(ticker, 'high'),
            'low': self.safe_string(ticker, 'low'),
            'bid': None,
            'bidVolume': None,
            'ask': None,
            'askVolume': None,
            'vwap': None,
            'open': None,
            'close': self.safe_string(ticker, 'last_price'),
            'last': self.safe_string(ticker, 'last_price'),
            'previousClose': None,
            'change': self.safe_string(ticker, 'price_change'),
            'percentage': self.safe_string(ticker, 'price_change_percentage'),
            'average': None,
            'baseVolume': None,
            'quoteVolume': self.safe_number(ticker, 'volume'),
            'info': ticker,
        }, market)

    async def watch_my_trades(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        :see: https://developers.bitpanda.com/exchange/#account-history-channel
        get the list of trades associated with the user
        :param str symbol: unified symbol of the market to fetch trades for. Use 'any' to watch all trades
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        await self.load_markets()
        messageHash = 'myTrades'
        if symbol is not None:
            market = self.market(symbol)
            symbol = market['symbol']
            messageHash += ':' + symbol
        await self.authenticate(params)
        url = self.urls['api']['ws']
        subscribeHash = 'ACCOUNT_HISTORY'
        bpRemainingQuota = self.safe_integer(self.options, 'bp_remaining_quota', 200)
        subscribe: dict = {
            'type': 'SUBSCRIBE',
            'bp_remaining_quota': bpRemainingQuota,
            'channels': [
                {
                    'name': 'ACCOUNT_HISTORY',
                },
            ],
        }
        request = self.deep_extend(subscribe, params)
        trades = await self.watch(url, messageHash, request, subscribeHash, request)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        trades = self.filter_by_symbol_since_limit(trades, symbol, since, limit)
        numTrades = len(trades)
        if numTrades == 0:
            return await self.watch_my_trades(symbol, since, limit, params)
        return trades

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        :see: https://developers.bitpanda.com/exchange/#market-ticker-channel
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        messageHash = 'book:' + symbol
        subscriptionHash = 'ORDER_BOOK'
        depth = 0
        if limit is not None:
            depth = limit
        request: dict = {
            'type': 'SUBSCRIBE',
            'channels': [
                {
                    'name': 'ORDER_BOOK',
                    'depth': depth,
                },
            ],
        }
        orderbook = await self.watch_many(messageHash, request, subscriptionHash, [symbol], params)
        return orderbook.limit()

    def handle_order_book(self, client: Client, message):
        #
        #  snapshot
        #     {
        #         "instrument_code": "ETH_BTC",
        #         "bids": [
        #             ['0.053595', "4.5352"],
        #             ...
        #         ],
        #         "asks": [
        #             ['0.055455', "0.2821"],
        #             ...
        #         ],
        #         "channel_name": "ORDER_BOOK",
        #         "type": "ORDER_BOOK_SNAPSHOT",
        #         "time": "2022-06-23T15:38:02.196282Z"
        #     }
        #
        #  update
        #     {
        #         "instrument_code": "ETH_BTC",
        #         "changes": [
        #             ["BUY", '0.053593', "8.0587"]
        #         ],
        #         "channel_name": "ORDER_BOOK",
        #         "type": "ORDER_BOOK_UPDATE",
        #         "time": "2022-06-23T15:38:02.751301Z"
        #     }
        #
        type = self.safe_string(message, 'type')
        marketId = self.safe_string(message, 'instrument_code')
        symbol = self.safe_symbol(marketId)
        dateTime = self.safe_string(message, 'time')
        timestamp = self.parse8601(dateTime)
        channel = 'book:' + symbol
        orderbook = self.safe_value(self.orderbooks, symbol)
        if orderbook is None:
            orderbook = self.order_book({})
        if type == 'ORDER_BOOK_SNAPSHOT':
            snapshot = self.parse_order_book(message, symbol, timestamp, 'bids', 'asks')
            orderbook.reset(snapshot)
        elif type == 'ORDER_BOOK_UPDATE':
            changes = self.safe_value(message, 'changes', [])
            self.handle_deltas(orderbook, changes)
        else:
            raise NotSupported(self.id + ' watchOrderBook() did not recognize message type ' + type)
        orderbook['nonce'] = timestamp
        orderbook['timestamp'] = timestamp
        orderbook['datetime'] = self.iso8601(timestamp)
        self.orderbooks[symbol] = orderbook
        client.resolve(orderbook, channel)

    def handle_delta(self, orderbook, delta):
        #
        #   ['BUY', "0.053595", "0"]
        #
        bidAsk = self.parse_bid_ask(delta, 1, 2)
        type = self.safe_string(delta, 0)
        if type == 'BUY':
            bids = orderbook['bids']
            bids.storeArray(bidAsk)
        elif type == 'SELL':
            asks = orderbook['asks']
            asks.storeArray(bidAsk)
        else:
            raise NotSupported(self.id + ' watchOrderBook() received unknown change type ' + self.json(delta))

    def handle_deltas(self, orderbook, deltas):
        #
        #    [
        #       ['BUY', "0.053593", "0"],
        #       ['SELL', "0.053698", "0"]
        #    ]
        #
        for i in range(0, len(deltas)):
            self.handle_delta(orderbook, deltas[i])

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        :see: https://developers.bitpanda.com/exchange/#account-history-channel
        watches information on multiple orders made by the user
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.channel]: can listen to orders using ACCOUNT_HISTORY or TRADING
        :returns dict[]: a list of `order structures <https://docs.ccxt.com/#/?id=order-structure>`
        """
        await self.load_markets()
        messageHash = 'orders'
        if symbol is not None:
            market = self.market(symbol)
            symbol = market['symbol']
            messageHash += ':' + symbol
        await self.authenticate(params)
        url = self.urls['api']['ws']
        subscribeHash = self.safe_string(params, 'channel', 'ACCOUNT_HISTORY')
        bpRemainingQuota = self.safe_integer(self.options, 'bp_remaining_quota', 200)
        subscribe: dict = {
            'type': 'SUBSCRIBE',
            'bp_remaining_quota': bpRemainingQuota,
            'channels': [
                {
                    'name': subscribeHash,
                },
            ],
        }
        request = self.deep_extend(subscribe, params)
        orders = await self.watch(url, messageHash, request, subscribeHash, request)
        if self.newUpdates:
            limit = orders.getLimit(symbol, limit)
        orders = self.filter_by_symbol_since_limit(orders, symbol, since, limit)
        numOrders = len(orders)
        if numOrders == 0:
            return await self.watch_orders(symbol, since, limit, params)
        return orders

    def handle_trading(self, client: Client, message):
        #
        #     {
        #         "order_book_sequence": 892925263,
        #         "side": "BUY",
        #         "amount": "0.00046",
        #         "trade_id": "d67b9b69-ab76-480f-9ba3-b33582202836",
        #         "matched_as": "TAKER",
        #         "matched_amount": "0.00046",
        #         "matched_price": "22231.08",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "7b39f316-0a71-4bfd-adda-3062e6f0bd37",
        #         "remaining": "0.0",
        #         "channel_name": "TRADING",
        #         "type": "FILL",
        #         "time": "2022-07-21T12:41:22.883341Z"
        #     }
        #
        #     {
        #         "status": "CANCELLED",
        #         "order_book_sequence": 892928424,
        #         "amount": "0.0003",
        #         "side": "SELL",
        #         "price": "50338.65",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "b3994a08-a9e8-4a79-a08b-33e3480382df",
        #         "remaining": "0.0003",
        #         "channel_name": "TRADING",
        #         "type": "DONE",
        #         "time": "2022-07-21T12:44:24.267000Z"
        #     }
        #
        #     {
        #         "order_book_sequence": 892934476,
        #         "side": "SELL",
        #         "amount": "0.00051",
        #         "price": "22349.02",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "1c6c585c-ec3d-4b94-9292-6c3d04a31dc8",
        #         "remaining": "0.00051",
        #         "channel_name": "TRADING",
        #         "type": "BOOKED",
        #         "time": "2022-07-21T12:50:10.093000Z"
        #     }
        #
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        order = self.parse_trading_order(message)
        orders = self.orders
        orders.append(order)
        client.resolve(self.orders, 'orders:' + order['symbol'])
        client.resolve(self.orders, 'orders')

    def parse_trading_order(self, order, market=None):
        #
        #     {
        #         "order_book_sequence": 892925263,
        #         "side": "BUY",
        #         "amount": "0.00046",
        #         "trade_id": "d67b9b69-ab76-480f-9ba3-b33582202836",
        #         "matched_as": "TAKER",
        #         "matched_amount": "0.00046",
        #         "matched_price": "22231.08",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "7b39f316-0a71-4bfd-adda-3062e6f0bd37",
        #         "remaining": "0.0",
        #         "channel_name": "TRADING",
        #         "type": "FILL",
        #         "time": "2022-07-21T12:41:22.883341Z"
        #     }
        #
        #     {
        #         "status": "CANCELLED",
        #         "order_book_sequence": 892928424,
        #         "amount": "0.0003",
        #         "side": "SELL",
        #         "price": "50338.65",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "b3994a08-a9e8-4a79-a08b-33e3480382df",
        #         "remaining": "0.0003",
        #         "channel_name": "TRADING",
        #         "type": "DONE",
        #         "time": "2022-07-21T12:44:24.267000Z"
        #     }
        #
        #     {
        #         "order_book_sequence": 892934476,
        #         "side": "SELL",
        #         "amount": "0.00051",
        #         "price": "22349.02",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "1c6c585c-ec3d-4b94-9292-6c3d04a31dc8",
        #         "remaining": "0.00051",
        #         "channel_name": "TRADING",
        #         "type": "BOOKED",
        #         "time": "2022-07-21T12:50:10.093000Z"
        #     }
        #
        #     {
        #         "type":"UPDATE",
        #         "channel_name": "TRADING",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "1e842f13-762a-4745-9f3b-07f1b43e7058",
        #         "client_id": "d75fb03b-b599-49e9-b926-3f0b6d103206",
        #         "time": "2020-01-11T01:01:01.999Z",
        #         "remaining": "1.23456",
        #         "order_book_sequence": 42,
        #         "status": "APPLIED",
        #         "amount": "1.35756",
        #         "amount_delta": "0.123",
        #         "modification_id": "cc0eed67-aecc-4fb4-a625-ff3890ceb4cc"
        #     }
        #  tracked
        #     {
        #         "type": "STOP_TRACKED",
        #         "channel_name": "TRADING",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "1e842f13-762a-4745-9f3b-07f1b43e7058",
        #         "client_id": "d75fb03b-b599-49e9-b926-3f0b6d103206",
        #         "time": "2020-01-11T01:01:01.999Z",
        #         "remaining": "1.23456",
        #         "order_book_sequence": 42,
        #         "trigger_price": "12345.67",
        #         "current_price": "11111.11"
        #     }
        #
        #     {
        #         "type": "STOP_TRIGGERED",
        #         "channel_name": "TRADING",
        #         "instrument_code": "BTC_EUR",
        #         "order_id": "1e842f13-762a-4745-9f3b-07f1b43e7058",
        #         "client_id": "d75fb03b-b599-49e9-b926-3f0b6d103206",
        #         "time": "2020-01-11T01:01:01.999Z",
        #         "remaining": "1.23456",
        #         "order_book_sequence": 42,
        #         "price": "13333.33"
        #     }
        #
        datetime = self.safe_string(order, 'time')
        marketId = self.safe_string(order, 'instrument_code')
        symbol = self.safe_symbol(marketId, market, '_')
        return self.safe_order({
            'id': self.safe_string(order, 'order_id'),
            'clientOrderId': self.safe_string(order, 'client_id'),
            'info': order,
            'timestamp': self.parse8601(datetime),
            'datetime': datetime,
            'lastTradeTimestamp': None,
            'symbol': symbol,
            'type': None,
            'timeInForce': None,
            'postOnly': None,
            'side': self.safe_string_lower(order, 'side'),
            'price': self.safe_number_2(order, 'price', 'matched_price'),
            'stopPrice': self.safe_number(order, 'trigger_price'),
            'amount': self.safe_number(order, 'amount'),
            'cost': None,
            'average': None,
            'filled': None,
            'remaining': self.safe_string(order, 'remaining'),
            'status': self.parse_trading_order_status(self.safe_string(order, 'status')),
            'fee': None,
            'trades': None,
        }, market)

    def parse_trading_order_status(self, status):
        statuses: dict = {
            'CANCELLED': 'canceled',
            'SELF_TRADE': 'rejected',
            'FILLED_FULLY': 'closed',
            'INSUFFICIENT_FUNDS': 'rejected',
            'INSUFFICIENT_LIQUIDITY': 'rejected',
            'TIME_TO_MARKET_EXCEEDED': 'rejected',
            'LAST_PRICE_UNKNOWN': 'rejected',
        }
        return self.safe_string(statuses, status, status)

    def handle_orders(self, client: Client, message):
        #
        #  snapshot
        #     {
        #         "account_id": "4920221a-48dc-423e-b336-bb65baccc7bd",
        #         "orders": [{
        #             "order": {
        #                 "order_id": "30e2de8f-9a34-472f-bcf8-3af4b7757626",
        #                 "account_holder": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #                 "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #                 "instrument_code": "BTC_EUR",
        #                 "time": "2022-06-28T06:10:02.587345Z",
        #                 "side": "SELL",
        #                 "price": "19645.48",
        #                 "amount": "0.00052",
        #                 "filled_amount": "0.00052",
        #                 "type": "MARKET",
        #                 "sequence": 7633339971,
        #                 "status": "FILLED_FULLY",
        #                 "average_price": "19645.48",
        #                 "is_post_only": False,
        #                 "order_book_sequence": 866885897,
        #                 "time_last_updated": "2022-06-28T06:10:02.766983Z",
        #                 "update_modification_sequence": 866885897
        #             },
        #             "trades": [{
        #                 "fee": {
        #                     "fee_amount": "0.01532347",
        #                     "fee_currency": "EUR",
        #                     "fee_percentage": "0.15",
        #                     "fee_group_id": "default",
        #                     "fee_type": "TAKER",
        #                     "running_trading_volume": "0.0",
        #                     "collection_type": "STANDARD"
        #                 },
        #                 "trade": {
        #                     "trade_id": "d83e302e-0b3a-4269-aa7d-ecf007cbe577",
        #                     "order_id": "30e2de8f-9a34-472f-bcf8-3af4b7757626",
        #                     "account_holder": "49203c1a-48dc-423e-b336-bb65baccc7bd",
        #                     "account_id": "4920221a-48dc-423e-b336-bb65baccc7bd",
        #                     "amount": "0.00052",
        #                     "side": "SELL",
        #                     "instrument_code": "BTC_EUR",
        #                     "price": "19645.48",
        #                     "time": "2022-06-28T06:10:02.693246Z",
        #                     "price_tick_sequence": 0,
        #                     "sequence": 7633339971
        #                 }
        #             }]
        #         }],
        #         "channel_name": "ACCOUNT_HISTORY",
        #         "type": "INACTIVE_ORDERS_SNAPSHOT",
        #         "time": "2022-06-28T06:11:52.469242Z"
        #     }
        #
        #
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        if self.myTrades is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            self.myTrades = ArrayCacheBySymbolById(limit)
        rawOrders = self.safe_value(message, 'orders', [])
        rawOrdersLength = len(rawOrders)
        if rawOrdersLength == 0:
            return
        orders = self.orders
        for i in range(0, len(rawOrders)):
            order = self.parse_order(rawOrders[i])
            symbol = self.safe_string(order, 'symbol', '')
            orders.append(order)
            client.resolve(self.orders, 'orders:' + symbol)
            rawTrades = self.safe_value(rawOrders[i], 'trades', [])
            for ii in range(0, len(rawTrades)):
                trade = self.parse_trade(rawTrades[ii])
                symbol = self.safe_string(trade, 'symbol', symbol)
                self.myTrades.append(trade)
                client.resolve(self.myTrades, 'myTrades:' + symbol)
        client.resolve(self.orders, 'orders')
        client.resolve(self.myTrades, 'myTrades')

    def handle_account_update(self, client: Client, message):
        #
        # order created
        #     {
        #         "account_id": "49302c1a-48dc-423e-b336-bb65baccc7bd",
        #         "sequence": 7658332018,
        #         "update": {
        #             "type": "ORDER_CREATED",
        #             "activity": "TRADING",
        #             "account_holder": "43202c1a-48dc-423e-b336-bb65baccc7bd",
        #             "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #             "order_id": "8893fd69-5ebd-496b-aaa4-269b4c18aa77",
        #             "time": "2022-06-29T04:33:29.661257Z",
        #             "order": {
        #                 "time_in_force": "GOOD_TILL_CANCELLED",
        #                 "is_post_only": False,
        #                 "order_id": "8892fd69-5ebd-496b-aaa4-269b4c18aa77",
        #                 "account_holder": "43202c1a-48dc-423e-b336-bb65baccc7bd",
        #                 "account_id": "49302c1a-48dc-423e-b336-bb65baccc7bd",
        #                 "instrument_code": "BTC_EUR",
        #                 "time": "2022-06-29T04:33:29.656896Z",
        #                 "side": "SELL",
        #                 "price": "50338.65",
        #                 "amount": "0.00021",
        #                 "filled_amount": "0.0",
        #                 "type": "LIMIT"
        #             },
        #             "locked": {
        #                 "currency_code": "BTC",
        #                 "amount": "0.00021",
        #                 "new_available": "0.00017",
        #                 "new_locked": "0.00021"
        #             },
        #             "id": "26e9c36a-b231-4bb0-a686-aa915a2fc9e6",
        #             "sequence": 7658332018
        #         },
        #         "channel_name": "ACCOUNT_HISTORY",
        #         "type": "ACCOUNT_UPDATE",
        #         "time": "2022-06-29T04:33:29.684517Z"
        #     }
        #
        #  order rejected
        #     {
        #         "account_id": "49302c1a-48dc-423e-b336-bb65baccc7bd",
        #         "sequence": 7658332018,
        #         "update": {
        #             "id": "d3fe6025-5b27-4df6-a957-98b8d131cb9d",
        #             "type": "ORDER_REJECTED",
        #             "activity": "TRADING",
        #             "account_id": "b355abb8-aaae-4fae-903c-c60ff74723c6",
        #             "sequence": 0,
        #             "timestamp": "2018-08-01T13:39:15.590Z",
        #             "reason": "INSUFFICIENT_FUNDS",
        #             "order_id": "6f991342-da2c-45c6-8830-8bf519cfc8cc",
        #             "client_id": "fb497387-8223-4111-87dc-66a86f98a7cf",
        #             "unlocked": {
        #                 "currency_code": "BTC",
        #                 "amount": "1.5",
        #                 "new_locked": "2.0",
        #                 "new_available": "1.5"
        #             }
        #         }
        #     }
        #
        #  order closed
        #     {
        #         "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #         "sequence": 7658471216,
        #         "update": {
        #             "type": "ORDER_CLOSED",
        #             "activity": "TRADING",
        #             "account_holder": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #             "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #             "time": "2022-06-29T04:43:57.169616Z",
        #             "order_id": "8892fd69-5ebd-496b-aaa4-269b4c18aa77",
        #             "unlocked": {
        #                 "currency_code": "BTC",
        #                 "amount": "0.00021",
        #                 "new_available": "0.00038",
        #                 "new_locked": "0.0"
        #             },
        #             "order_book_sequence": 867964191,
        #             "id": "26c5e1d7-65ba-4a11-a661-14c0130ff484",
        #             "sequence": 7658471216
        #         },
        #         "channel_name": "ACCOUNT_HISTORY",
        #         "type": "ACCOUNT_UPDATE",
        #         "time": "2022-06-29T04:43:57.182153Z"
        #     }
        #
        #  trade settled
        #     {
        #         "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #         "sequence": 7658502878,
        #         "update": {
        #             "type": "TRADE_SETTLED",
        #             "activity": "TRADING",
        #             "account_holder": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #             "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #             "time": "2022-06-29T04:46:12.933091Z",
        #             "order_id": "ad19951a-b616-401d-a062-8d0609f038a4",
        #             "order_book_sequence": 867965579,
        #             "filled_amount": "0.00052",
        #             "order": {
        #                 "amount": "0.00052",
        #                 "filled_amount": "0.00052"
        #             },
        #             "trade": {
        #                 "trade_id": "21039eb9-2df0-4227-be2d-0ea9b691ac66",
        #                 "order_id": "ad19951a-b616-401d-a062-8d0609f038a4",
        #                 "account_holder": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #                 "account_id": "49202c1a-48dc-423e-b336-bb65baccc7bd",
        #                 "amount": "0.00052",
        #                 "side": "BUY",
        #                 "instrument_code": "BTC_EUR",
        #                 "price": "19309.29",
        #                 "time": "2022-06-29T04:46:12.870581Z",
        #                 "price_tick_sequence": 0
        #             },
        #             "fee": {
        #                 "fee_amount": "0.00000078",
        #                 "fee_currency": "BTC",
        #                 "fee_percentage": "0.15",
        #                 "fee_group_id": "default",
        #                 "fee_type": "TAKER",
        #                 "running_trading_volume": "0.00052",
        #                 "collection_type": "STANDARD"
        #             },
        #             "spent": {
        #                 "currency_code": "EUR",
        #                 "amount": "10.0408308",
        #                 "new_available": "0.0",
        #                 "new_locked": "0.15949533"
        #             },
        #             "credited": {
        #                 "currency_code": "BTC",
        #                 "amount": "0.00051922",
        #                 "new_available": "0.00089922",
        #                 "new_locked": "0.0"
        #             },
        #             "unlocked": {
        #                 "currency_code": "EUR",
        #                 "amount": "0.0",
        #                 "new_available": "0.0",
        #                 "new_locked": "0.15949533"
        #             },
        #             "id": "22b40199-2508-4176-8a14-d4785c933444",
        #             "sequence": 7658502878
        #         },
        #         "channel_name": "ACCOUNT_HISTORY",
        #         "type": "ACCOUNT_UPDATE",
        #         "time": "2022-06-29T04:46:12.941837Z"
        #     }
        #
        #  Trade Settled with BEST fee collection enabled
        #     {
        #         "account_id": "49302c1a-48dc-423e-b336-bb65baccc7bd",
        #         "sequence": 7658951984,
        #         "update": {
        #             "id": "70e00504-d892-456f-9aae-4da7acb36aac",
        #             "sequence": 361792,
        #             "order_book_sequence": 123456,
        #             "type": "TRADE_SETTLED",
        #             "activity": "TRADING",
        #             "account_id": "379a12c0-4560-11e9-82fe-2b25c6f7d123",
        #             "time": "2019-10-22T12:09:55.731Z",
        #             "order_id": "9fcdd91c-7f6e-45f4-9956-61cddba55de5",
        #             "client_id": "fb497387-8223-4111-87dc-66a86f98a7cf",
        #             "order": {
        #                 "amount": "0.5",
        #                 "filled_amount": "0.5"
        #             },
        #             "trade": {
        #                 "trade_id": "a828b63e-b2cb-48f0-8d99-8fc22cf98e08",
        #                 "order_id": "9fcdd91c-7f6e-45f4-9956-61cddba55de5",
        #                 "account_id": "379a12c0-4560-11e9-82fe-2b25c6f7d123",
        #                 "amount": "0.5",
        #                 "side": "BUY",
        #                 "instrument_code": "BTC_EUR",
        #                 "price": "7451.6",
        #                 "time": "2019-10-22T12:09:55.667Z"
        #             },
        #             "fee": {
        #                 "fee_amount": "23.28625",
        #                 "fee_currency": "BEST",
        #                 "fee_percentage": "0.075",
        #                 "fee_group_id": "default",
        #                 "fee_type": "TAKER",
        #                 "running_trading_volume": "0.10058",
        #                 "collection_type": "BEST",
        #                 "applied_best_eur_rate": "1.04402"
        #             },
        #             "spent": {
        #                 "currency_code": "EUR",
        #                 "amount": "3725.8",
        #                 "new_available": "14517885.0675703028781",
        #                 "new_locked": "2354.882"
        #             },
        #             "spent_on_fees": {
        #                 "currency_code": "BEST",
        #                 "amount": "23.28625",
        #                 "new_available": "9157.31375",
        #                 "new_locked": "0.0"
        #             },
        #             "credited": {
        #                 "currency_code": "BTC",
        #                 "amount": "0.5",
        #                 "new_available": "5839.89633700481",
        #                 "new_locked": "0.0"
        #             },
        #             "unlocked": {
        #                 "currency_code": "EUR",
        #                 "amount": "0.15",
        #                 "new_available": "14517885.0675703028781",
        #                 "new_locked": "2354.882"
        #             }
        #         }
        #         "channel_name": "ACCOUNT_HISTORY",
        #         "type": "ACCOUNT_UPDATE",
        #         "time": "2022-06-29T05:18:51.760338Z"
        #     }
        #
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        if self.myTrades is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            self.myTrades = ArrayCacheBySymbolById(limit)
        symbol = None
        orders = self.orders
        update = self.safe_value(message, 'update', {})
        updateType = self.safe_string(update, 'type')
        if updateType == 'ORDER_REJECTED' or updateType == 'ORDER_CLOSED' or updateType == 'STOP_ORDER_TRIGGERED':
            orderId = self.safe_string(update, 'order_id')
            datetime = self.safe_string_2(update, 'time', 'timestamp')
            previousOrderArray = self.filter_by_array(self.orders, 'id', orderId, False)
            previousOrder = self.safe_value(previousOrderArray, 0, {})
            symbol = previousOrder['symbol']
            filled = self.safe_number(update, 'filled_amount')
            status = self.parse_ws_order_status(updateType)
            if updateType == 'ORDER_CLOSED' and filled == 0:
                status = 'canceled'
            orderObject: dict = {
                'id': orderId,
                'symbol': symbol,
                'status': status,
                'timestamp': self.parse8601(datetime),
                'datetime': datetime,
            }
            orders.append(orderObject)
        else:
            parsed = self.parse_order(update)
            symbol = self.safe_string(parsed, 'symbol', '')
            orders.append(parsed)
        client.resolve(self.orders, 'orders:' + symbol)
        client.resolve(self.orders, 'orders')
        # update balance
        balanceKeys = ['locked', 'unlocked', 'spent', 'spent_on_fees', 'credited', 'deducted']
        for i in range(0, len(balanceKeys)):
            newBalance = self.safe_value(update, balanceKeys[i])
            if newBalance is not None:
                self.update_balance(newBalance)
        client.resolve(self.balance, 'balance')
        # update trades
        if updateType == 'TRADE_SETTLED':
            parsed = self.parse_trade(update)
            symbol = self.safe_string(parsed, 'symbol', '')
            self.myTrades.append(parsed)
            client.resolve(self.myTrades, 'myTrades:' + symbol)
            client.resolve(self.myTrades, 'myTrades')

    def parse_ws_order_status(self, status):
        statuses: dict = {
            'ORDER_REJECTED': 'rejected',
            'ORDER_CLOSED': 'closed',
            'STOP_ORDER_TRIGGERED': 'triggered',
        }
        return self.safe_string(statuses, status, status)

    def update_balance(self, balance):
        #
        #     {
        #         "currency_code": "EUR",
        #         "amount": "0.0",
        #         "new_available": "0.0",
        #         "new_locked": "0.15949533"
        #     }
        #
        currencyId = self.safe_string(balance, 'currency_code')
        code = self.safe_currency_code(currencyId)
        account = self.account()
        account['free'] = self.safe_string(balance, 'new_available')
        account['used'] = self.safe_string(balance, 'new_locked')
        self.balance[code] = account
        self.balance = self.safe_balance(self.balance)

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        :see: https://developers.bitpanda.com/exchange/#candlesticks-channel
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        marketId = market['id']
        url = self.urls['api']['ws']
        timeframes = self.safe_value(self.options, 'timeframes', {})
        timeframeId = self.safe_value(timeframes, timeframe)
        if timeframeId is None:
            raise NotSupported(self.id + ' self interval is not supported, please provide one of the supported timeframes')
        messageHash = 'ohlcv.' + symbol + '.' + timeframe
        subscriptionHash = 'CANDLESTICKS'
        client = self.safe_value(self.clients, url)
        type = 'SUBSCRIBE'
        subscription = {}
        if client is not None:
            subscription = self.safe_value(client.subscriptions, subscriptionHash)
            if subscription is not None:
                ohlcvMarket = self.safe_value(subscription, marketId, {})
                marketSubscribed = self.safe_bool(ohlcvMarket, timeframe, False)
                if not marketSubscribed:
                    type = 'UPDATE_SUBSCRIPTION'
                    client.subscriptions[subscriptionHash] = None
            else:
                subscription = {}
        subscriptionMarketId = self.safe_value(subscription, marketId)
        if subscriptionMarketId is None:
            subscription[marketId] = {}
        subscription[marketId][timeframe] = True
        properties = []
        marketIds = list(subscription.keys())
        for i in range(0, len(marketIds)):
            marketIdtimeframes = list(subscription[marketIds[i]].keys())
            for ii in range(0, len(marketIdtimeframes)):
                marketTimeframeId = self.safe_value(timeframes, timeframe)
                property: dict = {
                    'instrument_code': marketIds[i],
                    'time_granularity': marketTimeframeId,
                }
                properties.append(property)
        request: dict = {
            'type': type,
            'channels': [
                {
                    'name': 'CANDLESTICKS',
                    'properties': properties,
                },
            ],
        }
        ohlcv = await self.watch(url, messageHash, self.deep_extend(request, params), subscriptionHash, subscription)
        if self.newUpdates:
            limit = ohlcv.getLimit(symbol, limit)
        return self.filter_by_since_limit(ohlcv, since, limit, 0, True)

    def handle_ohlcv(self, client: Client, message):
        #
        #  snapshot
        #     {
        #         "instrument_code": "BTC_EUR",
        #         "granularity": {unit: "MONTHS", period: 1},
        #         "high": "29750.81",
        #         "low": "16764.59",
        #         "open": "29556.02",
        #         "close": "20164.55",
        #         "volume": "107518944.610659",
        #         "last_sequence": 2275507,
        #         "channel_name": "CANDLESTICKS",
        #         "type": "CANDLESTICK_SNAPSHOT",
        #         "time": "2022-06-30T23:59:59.999000Z"
        #     }
        #
        #  update
        #     {
        #         "instrument_code": "BTC_EUR",
        #         "granularity": {
        #             "unit": "MINUTES",
        #             "period": 1
        #         },
        #         "high": "20164.16",
        #         "low": "20164.16",
        #         "open": "20164.16",
        #         "close": "20164.16",
        #         "volume": "3645.2768448",
        #         "last_sequence": 2275511,
        #         "channel_name": "CANDLESTICKS",
        #         "type": "CANDLESTICK",
        #         "time": "2022-06-24T21:20:59.999000Z"
        #     }
        #
        marketId = self.safe_string(message, 'instrument_code')
        symbol = self.safe_symbol(marketId)
        dateTime = self.safe_string(message, 'time')
        timeframeId = self.safe_value(message, 'granularity')
        timeframes = self.safe_value(self.options, 'timeframes', {})
        timeframe = self.find_timeframe(timeframeId, timeframes)
        channel = 'ohlcv.' + symbol + '.' + timeframe
        parsed = [
            self.parse8601(dateTime),
            self.safe_number(message, 'open'),
            self.safe_number(message, 'high'),
            self.safe_number(message, 'low'),
            self.safe_number(message, 'close'),
            self.safe_number(message, 'volume'),
        ]
        self.ohlcvs[symbol] = self.safe_value(self.ohlcvs, symbol, {})
        stored = self.safe_value(self.ohlcvs[symbol], timeframe)
        if stored is None:
            limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
            stored = ArrayCacheByTimestamp(limit)
        stored.append(parsed)
        self.ohlcvs[symbol][timeframe] = stored
        client.resolve(stored, channel)

    def find_timeframe(self, timeframe, timeframes=None):
        timeframes = timeframes or self.timeframes
        keys = list(timeframes.keys())
        for i in range(0, len(keys)):
            key = keys[i]
            if timeframes[key]['unit'] == timeframe['unit'] and timeframes[key]['period'] == timeframe['period']:
                return key
        return None

    def handle_subscriptions(self, client: Client, message):
        #
        #     {
        #         "channels": [{
        #             "instrument_codes": [Array],
        #             "depth": 0,
        #             "name": "ORDER_BOOK"
        #         }],
        #         "type": "SUBSCRIPTIONS",
        #         "time": "2022-06-23T15:36:26.948282Z"
        #     }
        #
        return message

    def handle_heartbeat(self, client: Client, message):
        #
        #     {
        #         "subscription": "SYSTEM",
        #         "channel_name": "SYSTEM",
        #         "type": "HEARTBEAT",
        #         "time": "2022-06-23T16:31:49.170224Z"
        #     }
        #
        return message

    def handle_error_message(self, client: Client, message):
        #
        #     {
        #         "error": "MALFORMED_JSON",
        #         "channel_name": "SYSTEM",
        #         "type": "ERROR",
        #         "time": "2022-06-23T15:38:25.470391Z"
        #     }
        #
        raise ExchangeError(self.id + ' ' + self.json(message))

    def handle_message(self, client: Client, message):
        error = self.safe_value(message, 'error')
        if error is not None:
            self.handle_error_message(client, message)
            return
        type = self.safe_value(message, 'type')
        handlers: dict = {
            'ORDER_BOOK_UPDATE': self.handle_order_book,
            'ORDER_BOOK_SNAPSHOT': self.handle_order_book,
            'ACTIVE_ORDERS_SNAPSHOT': self.handle_orders,
            'INACTIVE_ORDERS_SNAPSHOT': self.handle_orders,
            'ACCOUNT_UPDATE': self.handle_account_update,
            'BALANCES_SNAPSHOT': self.handle_balance_snapshot,
            'SUBSCRIPTIONS': self.handle_subscriptions,
            'SUBSCRIPTION_UPDATED': self.handle_subscriptions,
            'PRICE_TICK': self.handle_ticker,
            'PRICE_TICK_HISTORY': self.handle_subscriptions,
            'HEARTBEAT': self.handle_heartbeat,
            'MARKET_TICKER_UPDATES': self.handle_ticker,
            'PRICE_POINT_UPDATES': self.handle_price_point_updates,
            'CANDLESTICK_SNAPSHOT': self.handle_ohlcv,
            'CANDLESTICK': self.handle_ohlcv,
            'AUTHENTICATED': self.handle_authentication_message,
            'FILL': self.handle_trading,
            'DONE': self.handle_trading,
            'BOOKED': self.handle_trading,
            'UPDATE': self.handle_trading,
            'TRACKED': self.handle_trading,
            'TRIGGERED': self.handle_trading,
            'STOP_TRACKED': self.handle_trading,
            'STOP_TRIGGERED': self.handle_trading,
        }
        handler = self.safe_value(handlers, type)
        if handler is not None:
            handler(client, message)

    def handle_price_point_updates(self, client: Client, message):
        #
        #     {
        #         "channel_name": "MARKET_TICKER",
        #         "type": "PRICE_POINT_UPDATES",
        #         "time": "2019-03-01T10:59:59.999Z",
        #         "price_updates": [{
        #                 "instrument": "BTC_EUR",
        #                 "prices": [{
        #                         "time": "2019-03-01T08:59:59.999Z",
        #                         "close_price": "3580.6"
        #                     },
        #                     ...
        #                 ]
        #             },
        #             ...
        #         ]
        #     }
        #
        return message

    def handle_authentication_message(self, client: Client, message):
        #
        #    {
        #        "channel_name": "SYSTEM",
        #        "type": "AUTHENTICATED",
        #        "time": "2022-06-24T20:45:25.447488Z"
        #    }
        #
        future = self.safe_value(client.futures, 'authenticated')
        if future is not None:
            future.resolve(True)
        return message

    async def watch_many(self, messageHash, request, subscriptionHash, symbols: Strings = [], params={}):
        marketIds = []
        numSymbols = len(symbols)
        if numSymbols == 0:
            marketIds = list(self.markets_by_id.keys())
        else:
            marketIds = self.market_ids(symbols)
        url = self.urls['api']['ws']
        client = self.safe_value(self.clients, url)
        type = 'SUBSCRIBE'
        subscription = {}
        if client is not None:
            subscription = self.safe_value(client.subscriptions, subscriptionHash)
            if subscription is not None:
                for i in range(0, len(marketIds)):
                    marketId = marketIds[i]
                    marketSubscribed = self.safe_bool(subscription, marketId, False)
                    if not marketSubscribed:
                        type = 'UPDATE_SUBSCRIPTION'
                        client.subscriptions[subscriptionHash] = None
            else:
                subscription = {}
        for i in range(0, len(marketIds)):
            marketId = marketIds[i]
            subscription[marketId] = True
        request['type'] = type
        request['channels'][0]['instrument_codes'] = list(subscription.keys())
        return await self.watch(url, messageHash, self.deep_extend(request, params), subscriptionHash, subscription)

    async def authenticate(self, params={}):
        url = self.urls['api']['ws']
        client = self.client(url)
        messageHash = 'authenticated'
        future = client.future('authenticated')
        authenticated = self.safe_value(client.subscriptions, messageHash)
        if authenticated is None:
            self.check_required_credentials()
            request: dict = {
                'type': 'AUTHENTICATE',
                'api_token': self.apiKey,
            }
            self.watch(url, messageHash, self.extend(request, params), messageHash)
        return await future



import os
import sys

root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
sys.path.append(root)

# ----------------------------------------------------------------------------

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

# ----------------------------------------------------------------------------
# -*- coding: utf-8 -*-

from ccxt.test.base import test_shared_methods  # noqa E402

def test_margin_mode(exchange, skipped_properties, method, entry):
    format = {
        'info': {},
        'symbol': 'BTC/USDT:USDT',
        'marginMode': 'cross',
    }
    empty_allowed_for = ['symbol']
    test_shared_methods.assert_structure(exchange, skipped_properties, method, entry, format, empty_allowed_for)



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCache, ArrayCacheBySymbolById, ArrayCacheByTimestamp
from ccxt.base.types import Int, Market, Order, OrderBook, Str, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List
from ccxt.base.errors import ExchangeError


class hyperliquid(ccxt.async_support.hyperliquid):

    def describe(self):
        return self.deep_extend(super(hyperliquid, self).describe(), {
            'has': {
                'ws': True,
                'watchBalance': False,
                'watchMyTrades': True,
                'watchOHLCV': True,
                'watchOrderBook': True,
                'watchOrders': True,
                'watchTicker': False,
                'watchTickers': False,
                'watchTrades': True,
                'watchPosition': False,
            },
            'urls': {
                'api': {
                    'ws': {
                        'public': 'wss://api.hyperliquid.xyz/ws',
                    },
                },
                'test': {
                    'ws': {
                        'public': 'wss://api.hyperliquid-testnet.xyz/ws',
                    },
                },
            },
            'options': {
            },
            'streaming': {
                'ping': self.ping,
                'keepAlive': 20000,
            },
            'exceptions': {
                'ws': {
                    'exact': {
                    },
                },
            },
        })

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data
        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        messageHash = 'orderbook:' + symbol
        url = self.urls['api']['ws']['public']
        request: dict = {
            'method': 'subscribe',
            'subscription': {
                'type': 'l2Book',
                'coin': market['base'] if market['swap'] else market['id'],
            },
        }
        message = self.extend(request, params)
        orderbook = await self.watch(url, messageHash, message, messageHash)
        return orderbook.limit()

    def handle_order_book(self, client, message):
        #
        #     {
        #         "channel": "l2Book",
        #         "data": {
        #             "coin": "BTC",
        #             "time": 1710131872708,
        #             "levels": [
        #                 [
        #                     {
        #                         "px": "68674.0",
        #                         "sz": "0.97139",
        #                         "n": 4
        #                     }
        #                 ],
        #                 [
        #                     {
        #                         "px": "68675.0",
        #                         "sz": "0.04396",
        #                         "n": 1
        #                     }
        #                 ]
        #             ]
        #         }
        #     }
        #
        entry = self.safe_dict(message, 'data', {})
        coin = self.safe_string(entry, 'coin')
        marketId = self.coinToMarketId(coin)
        market = self.market(marketId)
        symbol = market['symbol']
        rawData = self.safe_list(entry, 'levels', [])
        data: dict = {
            'bids': self.safe_list(rawData, 0, []),
            'asks': self.safe_list(rawData, 1, []),
        }
        timestamp = self.safe_integer(entry, 'time')
        snapshot = self.parse_order_book(data, symbol, timestamp, 'bids', 'asks', 'px', 'sz')
        if not (symbol in self.orderbooks):
            ob = self.order_book(snapshot)
            self.orderbooks[symbol] = ob
        orderbook = self.orderbooks[symbol]
        orderbook.reset(snapshot)
        messageHash = 'orderbook:' + symbol
        client.resolve(orderbook, messageHash)

    async def watch_my_trades(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        watches information on multiple trades made by the user
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.user]: user address, will default to self.walletAddress if not provided
        :returns dict[]: a list of [order structures]{@link https://docs.ccxt.com/#/?id=order-structure
        """
        userAddress = None
        userAddress, params = self.handlePublicAddress('watchMyTrades', params)
        await self.load_markets()
        messageHash = 'myTrades'
        if symbol is not None:
            symbol = self.symbol(symbol)
            messageHash += ':' + symbol
        url = self.urls['api']['ws']['public']
        request: dict = {
            'method': 'subscribe',
            'subscription': {
                'type': 'userFills',
                'user': userAddress,
            },
        }
        message = self.extend(request, params)
        trades = await self.watch(url, messageHash, message, messageHash)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_symbol_since_limit(trades, symbol, since, limit, True)

    def handle_my_trades(self, client: Client, message):
        #
        #     {
        #         "channel": "userFills",
        #         "data": {
        #             "isSnapshot": True,
        #             "user": "0x15f43d1f2dee81424afd891943262aa90f22cc2a",
        #             "fills": [
        #                 {
        #                     "coin": "BTC",
        #                     "px": "72528.0",
        #                     "sz": "0.11693",
        #                     "side": "A",
        #                     "time": 1710208712815,
        #                     "startPosition": "0.11693",
        #                     "dir": "Close Long",
        #                     "closedPnl": "-0.81851",
        #                     "hash": "0xc5adaf35f8402750c218040b0a7bc301130051521273b6f398b3caad3e1f3f5f",
        #                     "oid": 7484888874,
        #                     "crossed": True,
        #                     "fee": "2.968244",
        #                     "liquidationMarkPx": null,
        #                     "tid": 567547935839686,
        #                     "cloid": null
        #                 }
        #             ]
        #         }
        #     }
        #
        entry = self.safe_dict(message, 'data', {})
        if self.myTrades is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            self.myTrades = ArrayCacheBySymbolById(limit)
        trades = self.myTrades
        symbols: dict = {}
        data = self.safe_list(entry, 'fills', [])
        dataLength = len(data)
        if dataLength == 0:
            return
        for i in range(0, len(data)):
            rawTrade = data[i]
            parsed = self.parse_ws_trade(rawTrade)
            symbol = parsed['symbol']
            symbols[symbol] = True
            trades.append(parsed)
        keys = list(symbols.keys())
        for i in range(0, len(keys)):
            currentMessageHash = 'myTrades:' + keys[i]
            client.resolve(trades, currentMessageHash)
        # non-symbol specific
        messageHash = 'myTrades'
        client.resolve(trades, messageHash)

    async def watch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        watches information on multiple trades made in a market
        :param str symbol: unified market symbol of the market trades were made in
        :param int [since]: the earliest time in ms to fetch trades for
        :param int [limit]: the maximum number of trade structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of [trade structures]{@link https://docs.ccxt.com/#/?id=trade-structure
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        messageHash = 'trade:' + symbol
        url = self.urls['api']['ws']['public']
        request: dict = {
            'method': 'subscribe',
            'subscription': {
                'type': 'trades',
                'coin': market['base'] if market['swap'] else market['id'],
            },
        }
        message = self.extend(request, params)
        trades = await self.watch(url, messageHash, message, messageHash)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp', True)

    def handle_trades(self, client: Client, message):
        #
        #     {
        #         "channel": "trades",
        #         "data": [
        #             {
        #                 "coin": "BTC",
        #                 "side": "A",
        #                 "px": "68517.0",
        #                 "sz": "0.005",
        #                 "time": 1710125266669,
        #                 "hash": "0xc872699f116e012186620407fc08a802015e0097c5cce74710697f7272e6e959",
        #                 "tid": 981894269203506
        #             }
        #         ]
        #     }
        #
        entry = self.safe_list(message, 'data', [])
        first = self.safe_dict(entry, 0, {})
        coin = self.safe_string(first, 'coin')
        marketId = self.coinToMarketId(coin)
        market = self.market(marketId)
        symbol = market['symbol']
        if not (symbol in self.trades):
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            stored = ArrayCache(limit)
            self.trades[symbol] = stored
        trades = self.trades[symbol]
        for i in range(0, len(entry)):
            data = self.safe_dict(entry, i)
            trade = self.parse_ws_trade(data)
            trades.append(trade)
        messageHash = 'trade:' + symbol
        client.resolve(trades, messageHash)

    def parse_ws_trade(self, trade, market: Market = None) -> Trade:
        #
        # fetchMyTrades
        #
        #     {
        #         "coin": "BTC",
        #         "px": "72528.0",
        #         "sz": "0.11693",
        #         "side": "A",
        #         "time": 1710208712815,
        #         "startPosition": "0.11693",
        #         "dir": "Close Long",
        #         "closedPnl": "-0.81851",
        #         "hash": "0xc5adaf35f8402750c218040b0a7bc301130051521273b6f398b3caad3e1f3f5f",
        #         "oid": 7484888874,
        #         "crossed": True,
        #         "fee": "2.968244",
        #         "liquidationMarkPx": null,
        #         "tid": 567547935839686,
        #         "cloid": null
        #     }
        #
        # fetchTrades
        #
        #     {
        #         "coin": "BTC",
        #         "side": "A",
        #         "px": "68517.0",
        #         "sz": "0.005",
        #         "time": 1710125266669,
        #         "hash": "0xc872699f116e012186620407fc08a802015e0097c5cce74710697f7272e6e959",
        #         "tid": 981894269203506
        #     }
        #
        timestamp = self.safe_integer(trade, 'time')
        price = self.safe_string(trade, 'px')
        amount = self.safe_string(trade, 'sz')
        coin = self.safe_string(trade, 'coin')
        marketId = self.coinToMarketId(coin)
        market = self.safe_market(marketId, None)
        symbol = market['symbol']
        id = self.safe_string(trade, 'tid')
        side = self.safe_string(trade, 'side')
        if side is not None:
            side = 'sell' if (side == 'A') else 'buy'
        fee = self.safe_string(trade, 'fee')
        return self.safe_trade({
            'info': trade,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': symbol,
            'id': id,
            'order': None,
            'type': None,
            'side': side,
            'takerOrMaker': None,
            'price': price,
            'amount': amount,
            'cost': None,
            'fee': {'cost': fee, 'currency': 'USDC'},
        }, market)

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        watches historical candlestick data containing the open, high, low, close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        market = self.market(symbol)
        symbol = market['symbol']
        url = self.urls['api']['ws']['public']
        request: dict = {
            'method': 'subscribe',
            'subscription': {
                'type': 'candle',
                'coin': market['base'] if market['swap'] else market['id'],
                'interval': timeframe,
            },
        }
        messageHash = 'candles:' + timeframe + ':' + symbol
        message = self.extend(request, params)
        ohlcv = await self.watch(url, messageHash, message, messageHash)
        if self.newUpdates:
            limit = ohlcv.getLimit(symbol, limit)
        return self.filter_by_since_limit(ohlcv, since, limit, 0, True)

    def handle_ohlcv(self, client: Client, message):
        #
        #     {
        #         channel: 'candle',
        #         data: {
        #             t: 1710146280000,
        #             T: 1710146339999,
        #             s: 'BTC',
        #             i: '1m',
        #             o: '71400.0',
        #             c: '71411.0',
        #             h: '71422.0',
        #             l: '71389.0',
        #             v: '1.20407',
        #             n: 20
        #         }
        #     }
        #
        data = self.safe_dict(message, 'data', {})
        base = self.safe_string(data, 's')
        marketId = self.coinToMarketId(base)
        symbol = self.safe_symbol(marketId)
        timeframe = self.safe_string(data, 'i')
        if not (symbol in self.ohlcvs):
            self.ohlcvs[symbol] = {}
        if not (timeframe in self.ohlcvs[symbol]):
            limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
            stored = ArrayCacheByTimestamp(limit)
            self.ohlcvs[symbol][timeframe] = stored
        ohlcv = self.ohlcvs[symbol][timeframe]
        parsed = self.parse_ohlcv(data)
        ohlcv.append(parsed)
        messageHash = 'candles:' + timeframe + ':' + symbol
        client.resolve(ohlcv, messageHash)

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        watches information on multiple orders made by the user
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.user]: user address, will default to self.walletAddress if not provided
        :returns dict[]: a list of [order structures]{@link https://docs.ccxt.com/#/?id=order-structure
        """
        await self.load_markets()
        userAddress = None
        userAddress, params = self.handlePublicAddress('watchOrders', params)
        market = None
        messageHash = 'order'
        if symbol is not None:
            market = self.market(symbol)
            symbol = market['symbol']
            messageHash = messageHash + ':' + symbol
        url = self.urls['api']['ws']['public']
        request: dict = {
            'method': 'subscribe',
            'subscription': {
                'type': 'orderUpdates',
                'user': userAddress,
            },
        }
        message = self.extend(request, params)
        orders = await self.watch(url, messageHash, message, messageHash)
        if self.newUpdates:
            limit = orders.getLimit(symbol, limit)
        return self.filter_by_symbol_since_limit(orders, symbol, since, limit, True)

    def handle_order(self, client: Client, message):
        #
        #     {
        #         channel: 'orderUpdates',
        #         data: [
        #             {
        #                 order: {
        #                     coin: 'BTC',
        #                     side: 'B',
        #                     limitPx: '30000.0',
        #                     sz: '0.001',
        #                     oid: 7456484275,
        #                     timestamp: 1710163596492,
        #                     origSz: '0.001'
        #                 },
        #                 status: 'open',
        #                 statusTimestamp: 1710163596492
        #             }
        #         ]
        #     }
        #
        data = self.safe_list(message, 'data', [])
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        dataLength = len(data)
        if dataLength == 0:
            return
        stored = self.orders
        messageHash = 'order'
        marketSymbols: dict = {}
        for i in range(0, len(data)):
            rawOrder = data[i]
            order = self.parse_order(rawOrder)
            stored.append(order)
            symbol = self.safe_string(order, 'symbol')
            marketSymbols[symbol] = True
        keys = list(marketSymbols.keys())
        for i in range(0, len(keys)):
            symbol = keys[i]
            innerMessageHash = messageHash + ':' + symbol
            client.resolve(stored, innerMessageHash)
        client.resolve(stored, messageHash)

    def handle_error_message(self, client: Client, message):
        #
        #     {
        #         "channel": "error",
        #         "data": "Error parsing JSON into valid websocket request: {\"type\": \"allMids\"}"
        #     }
        #
        channel = self.safe_string(message, 'channel', '')
        ret_msg = self.safe_string(message, 'data', '')
        if channel == 'error':
            raise ExchangeError(self.id + ' ' + ret_msg)
        else:
            return False

    def handle_message(self, client: Client, message):
        if self.handle_error_message(client, message):
            return
        topic = self.safe_string(message, 'channel', '')
        methods: dict = {
            'pong': self.handle_pong,
            'trades': self.handle_trades,
            'l2Book': self.handle_order_book,
            'candle': self.handle_ohlcv,
            'orderUpdates': self.handle_order,
            'userFills': self.handle_my_trades,
        }
        exacMethod = self.safe_value(methods, topic)
        if exacMethod is not None:
            exacMethod(client, message)
            return
        keys = list(methods.keys())
        for i in range(0, len(keys)):
            key = keys[i]
            if topic.find(keys[i]) >= 0:
                method = methods[key]
                method(client, message)
                return

    def ping(self, client):
        return {
            'method': 'ping',
        }

    def handle_pong(self, client: Client, message):
        #
        #   {
        #       "channel": "pong"
        #   }
        #
        client.lastPong = self.safe_integer(message, 'pong')
        return message



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_v1_market = publicGetV1Market = Entry('v1/market', 'public', 'GET', {'cost': 1})
    public_get_v2_udf_real_history = publicGetV2UdfRealHistory = Entry('v2/udf/real/history', 'public', 'GET', {'cost': 1})
    public_get_v1_market_id_order = publicGetV1MarketIdOrder = Entry('v1/market/{id}/order', 'public', 'GET', {'cost': 1})



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_book_currency_quote = publicGetBookCurrencyQuote = Entry('book/{currency}/{quote}', 'public', 'GET', {'cost': 1})
    public_get_chart_week = publicGetChartWeek = Entry('chart/week', 'public', 'GET', {'cost': 1})
    public_get_chart_week_currency_quote = publicGetChartWeekCurrencyQuote = Entry('chart/week/{currency}/{quote}', 'public', 'GET', {'cost': 1})
    public_get_currency = publicGetCurrency = Entry('currency', 'public', 'GET', {'cost': 1})
    public_get_currency_available = publicGetCurrencyAvailable = Entry('currency/available', 'public', 'GET', {'cost': 1})
    public_get_currency_quotes = publicGetCurrencyQuotes = Entry('currency/quotes', 'public', 'GET', {'cost': 1})
    public_get_currency_currency = publicGetCurrencyCurrency = Entry('currency/{currency}', 'public', 'GET', {'cost': 1})
    public_get_pair = publicGetPair = Entry('pair', 'public', 'GET', {'cost': 1})
    public_get_pair_available = publicGetPairAvailable = Entry('pair/available', 'public', 'GET', {'cost': 1})
    public_get_ticker = publicGetTicker = Entry('ticker', 'public', 'GET', {'cost': 1})
    public_get_ticker_base_quote = publicGetTickerBaseQuote = Entry('ticker/{base}/{quote}', 'public', 'GET', {'cost': 1})
    public_get_time = publicGetTime = Entry('time', 'public', 'GET', {'cost': 1})
    public_get_trade_history_currency_quote = publicGetTradeHistoryCurrencyQuote = Entry('trade/history/{currency}/{quote}', 'public', 'GET', {'cost': 1})
    public_get_trade_fee_currency_quote = publicGetTradeFeeCurrencyQuote = Entry('trade/fee/{currency}/{quote}', 'public', 'GET', {'cost': 1})
    public_get_trade_feelevels = publicGetTradeFeeLevels = Entry('trade/feeLevels', 'public', 'GET', {'cost': 1})
    public_get_transaction_bindings = publicGetTransactionBindings = Entry('transaction/bindings', 'public', 'GET', {'cost': 1})
    private_get_auth_account = privateGetAuthAccount = Entry('auth/account', 'private', 'GET', {'cost': 1})
    private_get_auth_account_currency_currency_type = privateGetAuthAccountCurrencyCurrencyType = Entry('auth/account/currency/{currency}/{type}', 'private', 'GET', {'cost': 1})
    private_get_auth_order = privateGetAuthOrder = Entry('auth/order', 'private', 'GET', {'cost': 1})
    private_get_auth_order_getorder_id = privateGetAuthOrderGetOrderId = Entry('auth/order/getOrder/{id}', 'private', 'GET', {'cost': 1})
    private_get_auth_order_pair_currency_quote = privateGetAuthOrderPairCurrencyQuote = Entry('auth/order/pair/{currency}/{quote}', 'private', 'GET', {'cost': 1})
    private_get_auth_order_pair_currency_quote_active = privateGetAuthOrderPairCurrencyQuoteActive = Entry('auth/order/pair/{currency}/{quote}/active', 'private', 'GET', {'cost': 1})
    private_get_auth_stoporder = privateGetAuthStopOrder = Entry('auth/stopOrder', 'private', 'GET', {'cost': 1})
    private_get_auth_stoporder_getorder_id = privateGetAuthStopOrderGetOrderId = Entry('auth/stopOrder/getOrder/{id}', 'private', 'GET', {'cost': 1})
    private_get_auth_stoporder_pair_currency_quote = privateGetAuthStopOrderPairCurrencyQuote = Entry('auth/stopOrder/pair/{currency}/{quote}', 'private', 'GET', {'cost': 1})
    private_get_auth_stoporder_pair_currency_quote_active = privateGetAuthStopOrderPairCurrencyQuoteActive = Entry('auth/stopOrder/pair/{currency}/{quote}/active', 'private', 'GET', {'cost': 1})
    private_get_auth_trade = privateGetAuthTrade = Entry('auth/trade', 'private', 'GET', {'cost': 1})
    private_get_auth_trade_pair_currency_quote = privateGetAuthTradePairCurrencyQuote = Entry('auth/trade/pair/{currency}/{quote}', 'private', 'GET', {'cost': 1})
    private_get_auth_trade_fee_currency_quote = privateGetAuthTradeFeeCurrencyQuote = Entry('auth/trade/fee/{currency}/{quote}', 'private', 'GET', {'cost': 1})
    private_get_auth_transaction = privateGetAuthTransaction = Entry('auth/transaction', 'private', 'GET', {'cost': 1})
    private_get_auth_transaction_bindings = privateGetAuthTransactionBindings = Entry('auth/transaction/bindings', 'private', 'GET', {'cost': 1})
    private_get_auth_transaction_bindings_currency = privateGetAuthTransactionBindingsCurrency = Entry('auth/transaction/bindings/{currency}', 'private', 'GET', {'cost': 1})
    private_get_auth_transaction_id = privateGetAuthTransactionId = Entry('auth/transaction/{id}', 'private', 'GET', {'cost': 1})
    private_get_auth_transfer = privateGetAuthTransfer = Entry('auth/transfer', 'private', 'GET', {'cost': 1})
    private_post_auth_order_cancel = privatePostAuthOrderCancel = Entry('auth/order/cancel', 'private', 'POST', {'cost': 1})
    private_post_auth_order_cancelall = privatePostAuthOrderCancelAll = Entry('auth/order/cancelAll', 'private', 'POST', {'cost': 1})
    private_post_auth_order_cancelall_currency_quote = privatePostAuthOrderCancelAllCurrencyQuote = Entry('auth/order/cancelAll/{currency}/{quote}', 'private', 'POST', {'cost': 1})
    private_post_auth_order_place = privatePostAuthOrderPlace = Entry('auth/order/place', 'private', 'POST', {'cost': 1})
    private_post_auth_spot_deposit = privatePostAuthSpotDeposit = Entry('auth/spot/deposit', 'private', 'POST', {'cost': 1})
    private_post_auth_spot_withdraw = privatePostAuthSpotWithdraw = Entry('auth/spot/withdraw', 'private', 'POST', {'cost': 1})
    private_post_auth_stoporder_cancel = privatePostAuthStopOrderCancel = Entry('auth/stopOrder/cancel', 'private', 'POST', {'cost': 1})
    private_post_auth_stoporder_cancelall = privatePostAuthStopOrderCancelAll = Entry('auth/stopOrder/cancelAll', 'private', 'POST', {'cost': 1})
    private_post_auth_stoporder_cancelall_currency_quote = privatePostAuthStopOrderCancelAllCurrencyQuote = Entry('auth/stopOrder/cancelAll/{currency}/{quote}', 'private', 'POST', {'cost': 1})
    private_post_auth_stoporder_place = privatePostAuthStopOrderPlace = Entry('auth/stopOrder/place', 'private', 'POST', {'cost': 1})
    private_post_auth_transaction_depositaddress = privatePostAuthTransactionDepositAddress = Entry('auth/transaction/depositAddress', 'private', 'POST', {'cost': 1})
    private_post_auth_transaction_withdraw = privatePostAuthTransactionWithdraw = Entry('auth/transaction/withdraw', 'private', 'POST', {'cost': 1})
    private_post_auth_transaction_withdraw_cancel = privatePostAuthTransactionWithdrawCancel = Entry('auth/transaction/withdraw/cancel', 'private', 'POST', {'cost': 1})
    private_post_auth_transaction_withdraw_confirm = privatePostAuthTransactionWithdrawConfirm = Entry('auth/transaction/withdraw/confirm', 'private', 'POST', {'cost': 1})
    private_post_auth_transaction_withdraw_resendcode = privatePostAuthTransactionWithdrawResendCode = Entry('auth/transaction/withdraw/resendCode', 'private', 'POST', {'cost': 1})
    private_post_auth_transfer_email = privatePostAuthTransferEmail = Entry('auth/transfer/email', 'private', 'POST', {'cost': 1})
    private_post_auth_transfer_id = privatePostAuthTransferId = Entry('auth/transfer/id', 'private', 'POST', {'cost': 1})
    private_post_auth_transfer_phone = privatePostAuthTransferPhone = Entry('auth/transfer/phone', 'private', 'POST', {'cost': 1})



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

from ccxt.base.exchange import Exchange
from ccxt.abstract.eterex import ImplicitAPI
from ccxt.base.types import Market, Strings, Ticker, Tickers
from typing import List


class eterex(Exchange, ImplicitAPI):

    def describe(self):
        return self.deep_extend(super(eterex, self).describe(), {
            'id': 'eterex',
            'name': 'Eterex',
            'country': ['IR'],
            'rateLimit': 1000,
            'version': '1',
            'certified': False,
            'pro': False,
            'has': {
                'CORS': None,
                'spot': True,
                'margin': False,
                'swap': False,
                'future': False,
                'option': False,
                'addMargin': False,
                'cancelAllOrders': False,
                'cancelOrder': False,
                'cancelOrders': False,
                'createDepositAddress': False,
                'createOrder': False,
                'createStopLimitOrder': False,
                'createStopMarketOrder': False,
                'createStopOrder': False,
                'editOrder': False,
                'fetchBalance': False,
                'fetchBorrowInterest': False,
                'fetchBorrowRateHistories': False,
                'fetchBorrowRateHistory': False,
                'fetchClosedOrders': False,
                'fetchCrossBorrowRate': False,
                'fetchCrossBorrowRates': False,
                'fetchCurrencies': False,
                'fetchDepositAddress': False,
                'fetchDeposits': False,
                'fetchFundingHistory': False,
                'fetchFundingRate': False,
                'fetchFundingRateHistory': False,
                'fetchFundingRates': False,
                'fetchIndexOHLCV': False,
                'fetchIsolatedBorrowRate': False,
                'fetchIsolatedBorrowRates': False,
                'fetchL2OrderBook': False,
                'fetchL3OrderBook': False,
                'fetchLedger': False,
                'fetchLedgerEntry': False,
                'fetchLeverageTiers': False,
                'fetchMarkets': True,
                'fetchMarkOHLCV': False,
                'fetchMyTrades': False,
                'fetchOHLCV': False,
                'fetchOpenInterestHistory': False,
                'fetchOpenOrders': False,
                'fetchOrder': False,
                'fetchOrderBook': False,
                'fetchOrders': False,
                'fetchOrderTrades': 'emulated',
                'fetchPositions': False,
                'fetchPremiumIndexOHLCV': False,
                'fetchTicker': True,
                'fetchTickers': True,
                'fetchTime': False,
                'fetchTrades': False,
                'fetchTradingFee': False,
                'fetchTradingFees': False,
                'fetchWithdrawals': False,
                'setLeverage': False,
                'setMarginMode': False,
                'transfer': False,
                'withdraw': False,
            },
            'comment': 'This comment is optional',
            'urls': {
                'logo': 'https://cdn.arz.digital/cr-odin/img/exchanges/eterex/64x64.png',
                'api': {
                    'public': 'https://reward.eterex.com',
                },
                'www': 'https://eterex.com',
                'doc': [
                    'https://eterex.com',
                ],
            },
            'api': {
                'public': {
                    'get': {
                        'ad/rates': 1,
                    },
                },
            },
            'fees': {
                'trading': {
                    'tierBased': False,
                    'percentage': True,
                    'maker': self.parse_number('0.001'),
                    'taker': self.parse_number('0.001'),
                },
            },
        })

    def fetch_markets(self, symbols: Strings = None, params={}) -> List[Market]:
        """
        retrieves data on all markets for eterex
        :see: https://reward.eterex.com
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: an array of objects representing market data
        """
        response = self.publicGetAdRates(params)
        markets = self.safe_dict(response, 'markets')
        marketKeys = list(markets.keys())
        result = []
        for i in range(0, len(marketKeys)):
            index = marketKeys[i]
            market = self.parse_market(markets[index])
            result.append(market)
        return result

    def parse_market(self, market) -> Market:
        #  {
        #     'sell_price': 4001252619,
        #     'buy_price': 3960630765,
        #     'base': 'BTC',
        #     'quote': 'IRT',
        #     'sell_max': 9000,
        #     'buy_max': 9000,
        #     'sell_active': True,
        #     'buy_active': True,
        #     '24h_volume': 44734409872.22664,
        # },
        baseId = self.safe_string(market, 'base')
        quoteId = self.safe_string(market, 'quote')
        base = self.safe_currency_code(baseId)
        quote = self.safe_currency_code(quoteId)
        id = base + quote
        baseId = baseId.lower()
        quoteId = quoteId.lower()
        return {
            'id': id,
            'symbol': base + '/' + quote,
            'base': base,
            'quote': quote,
            'settle': None,
            'baseId': baseId,
            'quoteId': quoteId,
            'settleId': None,
            'type': 'spot',
            'spot': True,
            'margin': False,
            'swap': False,
            'future': False,
            'option': False,
            'active': True,
            'contract': False,
            'linear': None,
            'inverse': None,
            'contractSize': None,
            'expiry': None,
            'expiryDatetime': None,
            'strike': None,
            'optionType': None,
            'precision': {
                'amount': None,
                'price': None,
            },
            'limits': {
                'leverage': {
                    'min': None,
                    'max': None,
                },
                'amount': {
                    'min': None,
                    'max': None,
                },
                'price': {
                    'min': None,
                    'max': None,
                },
                'cost': {
                    'min': None,
                    'max': None,
                },
            },
            'created': None,
            'info': market,
        }

    def fetch_tickers(self, symbols: Strings = None, params={}) -> Tickers:
        """
        fetches price tickers for multiple markets, statistical information calculated over the past 24 hours for each market
        :see: https://reward.eterex.com
        :param str[]|None symbols: unified symbols of the markets to fetch the ticker for, all market tickers are returned if not assigned
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a dictionary of `ticker structures <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        self.load_markets()
        if symbols is not None:
            symbols = self.market_symbols(symbols)
        response = self.publicGetAdRates(params)
        markets = self.safe_dict(response, 'markets')
        marketKeys = list(markets.keys())
        result = {}
        for i in range(0, len(marketKeys)):
            index = marketKeys[i]
            markets[index]['symbol'] = index
            ticker = self.parse_ticker(markets[index])
            symbol = ticker['symbol']
            result[symbol] = ticker
        return self.filter_by_array_tickers(result, 'symbol', symbols)

    def fetch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        fetches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market
        :see: https://reward.eterex.com
        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        self.load_markets()
        market = self.market(symbol)
        response = self.publicGetAdRates(params)
        markets = self.safe_dict(response, 'markets')
        markets[market['id']]['symbol'] = market['id']
        ticker = self.parse_ticker(markets[market['id']])
        return ticker

    def parse_ticker(self, ticker, market: Market = None) -> Ticker:
        #  {
        #     'sell_price': 4001252619,
        #     'buy_price': 3960630765,
        #     'base': 'BTC',
        #     'quote': 'IRT',
        #     'sell_max': 9000,
        #     'buy_max': 9000,
        #     'sell_active': True,
        #     'buy_active': True,
        #     '24h_volume': 44734409872.22664,
        # },
        marketType = 'otc'
        marketId = self.safe_string(ticker, 'symbol')
        symbol = self.safe_symbol(marketId, market, None, marketType)
        high = self.safe_float(ticker, 'sell_price', 0)
        low = self.safe_float(ticker, 'buy_price', 0)
        bid = self.safe_float(ticker, 'sell_price', 0)
        ask = self.safe_float(ticker, 'buy_price', 0)
        last = self.safe_float(ticker, 'buy_price', 0)
        quoteVolume = self.safe_float(ticker, '24h_volume', 0)
        return self.safe_ticker({
            'symbol': symbol,
            'timestamp': None,
            'datetime': None,
            'high': high,
            'low': low,
            'bid': bid,
            'bidVolume': None,
            'ask': ask,
            'askVolume': None,
            'vwap': None,
            'open': None,
            'close': last,
            'last': last,
            'previousClose': None,
            'change': None,
            'percentage': None,
            'average': None,
            'baseVolume': None,
            'quoteVolume': quoteVolume,
            'info': ticker,
        }, market)

    def sign(self, path, api='public', method='GET', params={}, headers=None, body=None):
        url = self.urls['api']['public'] + '/' + path
        headers = {'Content-Type': 'application/json'}
        return {'url': url, 'method': method, 'body': body, 'headers': headers}



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_countries = publicGetCountries = Entry('countries', 'public', 'GET', {})
    public_get_currencies = publicGetCurrencies = Entry('currencies', 'public', 'GET', {})
    public_get_data_currency_ticker = publicGetDataCurrencyTicker = Entry('data/{currency}/ticker', 'public', 'GET', {})
    public_get_data_currency_trades = publicGetDataCurrencyTrades = Entry('data/{currency}/trades', 'public', 'GET', {})
    public_get_data_currency_depth = publicGetDataCurrencyDepth = Entry('data/{currency}/depth', 'public', 'GET', {})
    public_get_bitcoin_charts_id_trades = publicGetBitcoinChartsIdTrades = Entry('bitcoin_charts/{id}/trades', 'public', 'GET', {})
    public_get_bitcoin_charts_id_depth = publicGetBitcoinChartsIdDepth = Entry('bitcoin_charts/{id}/depth', 'public', 'GET', {})
    private_get_user = privateGetUser = Entry('user', 'private', 'GET', {})
    private_get_user_addresses = privateGetUserAddresses = Entry('user/addresses', 'private', 'GET', {})
    private_get_user_addresses_address = privateGetUserAddressesAddress = Entry('user/addresses/{address}', 'private', 'GET', {})
    private_get_user_orders = privateGetUserOrders = Entry('user/orders', 'private', 'GET', {})
    private_get_user_orders_uuid = privateGetUserOrdersUuid = Entry('user/orders/{uuid}', 'private', 'GET', {})
    private_get_user_price_alerts = privateGetUserPriceAlerts = Entry('user/price_alerts', 'private', 'GET', {})
    private_get_merchant_get_payment_uuid = privateGetMerchantGetPaymentUuid = Entry('merchant/get_payment/{uuid}', 'private', 'GET', {})
    private_post_user_addresses = privatePostUserAddresses = Entry('user/addresses', 'private', 'POST', {})
    private_post_user_orders = privatePostUserOrders = Entry('user/orders', 'private', 'POST', {})
    private_post_user_withdrawals = privatePostUserWithdrawals = Entry('user/withdrawals', 'private', 'POST', {})
    private_post_user_email_transfers = privatePostUserEmailTransfers = Entry('user/email_transfers', 'private', 'POST', {})
    private_post_user_payment_requests = privatePostUserPaymentRequests = Entry('user/payment_requests', 'private', 'POST', {})
    private_post_user_price_alerts = privatePostUserPriceAlerts = Entry('user/price_alerts', 'private', 'POST', {})
    private_post_merchant_create_payment = privatePostMerchantCreatePayment = Entry('merchant/create_payment', 'private', 'POST', {})
    private_delete_user_orders_uuid = privateDeleteUserOrdersUuid = Entry('user/orders/{uuid}', 'private', 'DELETE', {})
    private_delete_user_orders_uuid_cancel = privateDeleteUserOrdersUuidCancel = Entry('user/orders/{uuid}/cancel', 'private', 'DELETE', {})
    private_delete_user_price_alerts_id = privateDeleteUserPriceAlertsId = Entry('user/price_alerts/{id}', 'private', 'DELETE', {})



import os
import sys

root = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
sys.path.append(root)

# ----------------------------------------------------------------------------

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

# ----------------------------------------------------------------------------
# -*- coding: utf-8 -*-

from ccxt.test.base import test_shared_methods  # noqa E402

def test_ohlcv(exchange, skipped_properties, method, entry, symbol, now):
    format = [1638230400000, exchange.parse_number('0.123'), exchange.parse_number('0.125'), exchange.parse_number('0.121'), exchange.parse_number('0.122'), exchange.parse_number('123.456')]
    empty_not_allowed_for = [0, 1, 2, 3, 4, 5]
    test_shared_methods.assert_structure(exchange, skipped_properties, method, entry, format, empty_not_allowed_for)
    test_shared_methods.assert_timestamp_and_datetime(exchange, skipped_properties, method, entry, now, 0)
    log_text = test_shared_methods.log_template(exchange, method, entry)
    #
    assert len(entry) >= 6, 'ohlcv array length should be >= 6;' + log_text
    if not ('roundTimestamp' in skipped_properties):
        test_shared_methods.assert_round_minute_timestamp(exchange, skipped_properties, method, entry, 0)
    high = exchange.safe_string(entry, 2)
    low = exchange.safe_string(entry, 3)
    test_shared_methods.assert_less_or_equal(exchange, skipped_properties, method, entry, '1', high)
    test_shared_methods.assert_greater_or_equal(exchange, skipped_properties, method, entry, '1', low)
    test_shared_methods.assert_less_or_equal(exchange, skipped_properties, method, entry, '4', high)
    test_shared_methods.assert_greater_or_equal(exchange, skipped_properties, method, entry, '4', low)
    assert (symbol is None) or (isinstance(symbol, str)), 'symbol ' + symbol + ' is incorrect' + log_text  # todo: check with standard symbol check



from ccxt.base.types import Entry


class ImplicitAPI:
    web_get_v1_healthcheck = webGetV1Healthcheck = Entry('v1/healthcheck', 'web', 'GET', {})
    v1_public_get_markets = v1PublicGetMarkets = Entry('markets', ['v1', 'public'], 'GET', {})
    v1_public_get_tickers = v1PublicGetTickers = Entry('tickers', ['v1', 'public'], 'GET', {})
    v1_public_get_ticker = v1PublicGetTicker = Entry('ticker', ['v1', 'public'], 'GET', {})
    v1_public_get_symbols = v1PublicGetSymbols = Entry('symbols', ['v1', 'public'], 'GET', {})
    v1_public_get_depth_result = v1PublicGetDepthResult = Entry('depth/result', ['v1', 'public'], 'GET', {})
    v1_public_get_history = v1PublicGetHistory = Entry('history', ['v1', 'public'], 'GET', {})
    v1_public_get_kline = v1PublicGetKline = Entry('kline', ['v1', 'public'], 'GET', {})
    v1_private_post_account_balance = v1PrivatePostAccountBalance = Entry('account/balance', ['v1', 'private'], 'POST', {})
    v1_private_post_order_new = v1PrivatePostOrderNew = Entry('order/new', ['v1', 'private'], 'POST', {})
    v1_private_post_order_cancel = v1PrivatePostOrderCancel = Entry('order/cancel', ['v1', 'private'], 'POST', {})
    v1_private_post_orders = v1PrivatePostOrders = Entry('orders', ['v1', 'private'], 'POST', {})
    v1_private_post_account_order_history = v1PrivatePostAccountOrderHistory = Entry('account/order_history', ['v1', 'private'], 'POST', {})
    v1_private_post_account_executed_history = v1PrivatePostAccountExecutedHistory = Entry('account/executed_history', ['v1', 'private'], 'POST', {})
    v1_private_post_account_executed_history_all = v1PrivatePostAccountExecutedHistoryAll = Entry('account/executed_history/all', ['v1', 'private'], 'POST', {})
    v1_private_post_account_order = v1PrivatePostAccountOrder = Entry('account/order', ['v1', 'private'], 'POST', {})
    v2_public_get_markets = v2PublicGetMarkets = Entry('markets', ['v2', 'public'], 'GET', {})
    v2_public_get_ticker = v2PublicGetTicker = Entry('ticker', ['v2', 'public'], 'GET', {})
    v2_public_get_assets = v2PublicGetAssets = Entry('assets', ['v2', 'public'], 'GET', {})
    v2_public_get_fee = v2PublicGetFee = Entry('fee', ['v2', 'public'], 'GET', {})
    v2_public_get_depth_market = v2PublicGetDepthMarket = Entry('depth/{market}', ['v2', 'public'], 'GET', {})
    v2_public_get_trades_market = v2PublicGetTradesMarket = Entry('trades/{market}', ['v2', 'public'], 'GET', {})
    v4_public_get_assets = v4PublicGetAssets = Entry('assets', ['v4', 'public'], 'GET', {})
    v4_public_get_collateral_markets = v4PublicGetCollateralMarkets = Entry('collateral/markets', ['v4', 'public'], 'GET', {})
    v4_public_get_fee = v4PublicGetFee = Entry('fee', ['v4', 'public'], 'GET', {})
    v4_public_get_orderbook_market = v4PublicGetOrderbookMarket = Entry('orderbook/{market}', ['v4', 'public'], 'GET', {})
    v4_public_get_ticker = v4PublicGetTicker = Entry('ticker', ['v4', 'public'], 'GET', {})
    v4_public_get_trades_market = v4PublicGetTradesMarket = Entry('trades/{market}', ['v4', 'public'], 'GET', {})
    v4_public_get_time = v4PublicGetTime = Entry('time', ['v4', 'public'], 'GET', {})
    v4_public_get_ping = v4PublicGetPing = Entry('ping', ['v4', 'public'], 'GET', {})
    v4_public_get_markets = v4PublicGetMarkets = Entry('markets', ['v4', 'public'], 'GET', {})
    v4_public_get_futures = v4PublicGetFutures = Entry('futures', ['v4', 'public'], 'GET', {})
    v4_public_get_platform_status = v4PublicGetPlatformStatus = Entry('platform/status', ['v4', 'public'], 'GET', {})
    v4_private_post_collateral_account_balance = v4PrivatePostCollateralAccountBalance = Entry('collateral-account/balance', ['v4', 'private'], 'POST', {})
    v4_private_post_collateral_account_balance_summary = v4PrivatePostCollateralAccountBalanceSummary = Entry('collateral-account/balance-summary', ['v4', 'private'], 'POST', {})
    v4_private_post_collateral_account_positions_history = v4PrivatePostCollateralAccountPositionsHistory = Entry('collateral-account/positions/history', ['v4', 'private'], 'POST', {})
    v4_private_post_collateral_account_leverage = v4PrivatePostCollateralAccountLeverage = Entry('collateral-account/leverage', ['v4', 'private'], 'POST', {})
    v4_private_post_collateral_account_positions_open = v4PrivatePostCollateralAccountPositionsOpen = Entry('collateral-account/positions/open', ['v4', 'private'], 'POST', {})
    v4_private_post_collateral_account_summary = v4PrivatePostCollateralAccountSummary = Entry('collateral-account/summary', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_address = v4PrivatePostMainAccountAddress = Entry('main-account/address', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_balance = v4PrivatePostMainAccountBalance = Entry('main-account/balance', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_create_new_address = v4PrivatePostMainAccountCreateNewAddress = Entry('main-account/create-new-address', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_codes = v4PrivatePostMainAccountCodes = Entry('main-account/codes', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_codes_apply = v4PrivatePostMainAccountCodesApply = Entry('main-account/codes/apply', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_codes_my = v4PrivatePostMainAccountCodesMy = Entry('main-account/codes/my', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_codes_history = v4PrivatePostMainAccountCodesHistory = Entry('main-account/codes/history', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_fiat_deposit_url = v4PrivatePostMainAccountFiatDepositUrl = Entry('main-account/fiat-deposit-url', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_history = v4PrivatePostMainAccountHistory = Entry('main-account/history', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_withdraw = v4PrivatePostMainAccountWithdraw = Entry('main-account/withdraw', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_withdraw_pay = v4PrivatePostMainAccountWithdrawPay = Entry('main-account/withdraw-pay', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_transfer = v4PrivatePostMainAccountTransfer = Entry('main-account/transfer', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_smart_plans = v4PrivatePostMainAccountSmartPlans = Entry('main-account/smart/plans', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_smart_investment = v4PrivatePostMainAccountSmartInvestment = Entry('main-account/smart/investment', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_smart_investment_close = v4PrivatePostMainAccountSmartInvestmentClose = Entry('main-account/smart/investment/close', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_smart_investments = v4PrivatePostMainAccountSmartInvestments = Entry('main-account/smart/investments', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_fee = v4PrivatePostMainAccountFee = Entry('main-account/fee', ['v4', 'private'], 'POST', {})
    v4_private_post_main_account_smart_interest_payment_history = v4PrivatePostMainAccountSmartInterestPaymentHistory = Entry('main-account/smart/interest-payment-history', ['v4', 'private'], 'POST', {})
    v4_private_post_trade_account_balance = v4PrivatePostTradeAccountBalance = Entry('trade-account/balance', ['v4', 'private'], 'POST', {})
    v4_private_post_trade_account_executed_history = v4PrivatePostTradeAccountExecutedHistory = Entry('trade-account/executed-history', ['v4', 'private'], 'POST', {})
    v4_private_post_trade_account_order = v4PrivatePostTradeAccountOrder = Entry('trade-account/order', ['v4', 'private'], 'POST', {})
    v4_private_post_trade_account_order_history = v4PrivatePostTradeAccountOrderHistory = Entry('trade-account/order/history', ['v4', 'private'], 'POST', {})
    v4_private_post_order_collateral_limit = v4PrivatePostOrderCollateralLimit = Entry('order/collateral/limit', ['v4', 'private'], 'POST', {})
    v4_private_post_order_collateral_market = v4PrivatePostOrderCollateralMarket = Entry('order/collateral/market', ['v4', 'private'], 'POST', {})
    v4_private_post_order_collateral_stop_limit = v4PrivatePostOrderCollateralStopLimit = Entry('order/collateral/stop-limit', ['v4', 'private'], 'POST', {})
    v4_private_post_order_collateral_trigger_market = v4PrivatePostOrderCollateralTriggerMarket = Entry('order/collateral/trigger-market', ['v4', 'private'], 'POST', {})
    v4_private_post_order_new = v4PrivatePostOrderNew = Entry('order/new', ['v4', 'private'], 'POST', {})
    v4_private_post_order_market = v4PrivatePostOrderMarket = Entry('order/market', ['v4', 'private'], 'POST', {})
    v4_private_post_order_stock_market = v4PrivatePostOrderStockMarket = Entry('order/stock_market', ['v4', 'private'], 'POST', {})
    v4_private_post_order_stop_limit = v4PrivatePostOrderStopLimit = Entry('order/stop_limit', ['v4', 'private'], 'POST', {})
    v4_private_post_order_stop_market = v4PrivatePostOrderStopMarket = Entry('order/stop_market', ['v4', 'private'], 'POST', {})
    v4_private_post_order_cancel = v4PrivatePostOrderCancel = Entry('order/cancel', ['v4', 'private'], 'POST', {})
    v4_private_post_order_cancel_all = v4PrivatePostOrderCancelAll = Entry('order/cancel/all', ['v4', 'private'], 'POST', {})
    v4_private_post_order_kill_switch = v4PrivatePostOrderKillSwitch = Entry('order/kill-switch', ['v4', 'private'], 'POST', {})
    v4_private_post_order_kill_switch_status = v4PrivatePostOrderKillSwitchStatus = Entry('order/kill-switch/status', ['v4', 'private'], 'POST', {})
    v4_private_post_order_bulk = v4PrivatePostOrderBulk = Entry('order/bulk', ['v4', 'private'], 'POST', {})
    v4_private_post_order_modify = v4PrivatePostOrderModify = Entry('order/modify', ['v4', 'private'], 'POST', {})
    v4_private_post_orders = v4PrivatePostOrders = Entry('orders', ['v4', 'private'], 'POST', {})
    v4_private_post_oco_orders = v4PrivatePostOcoOrders = Entry('oco-orders', ['v4', 'private'], 'POST', {})
    v4_private_post_order_collateral_oco = v4PrivatePostOrderCollateralOco = Entry('order/collateral/oco', ['v4', 'private'], 'POST', {})
    v4_private_post_order_oco_cancel = v4PrivatePostOrderOcoCancel = Entry('order/oco-cancel', ['v4', 'private'], 'POST', {})
    v4_private_post_order_oto_cancel = v4PrivatePostOrderOtoCancel = Entry('order/oto-cancel', ['v4', 'private'], 'POST', {})
    v4_private_post_profile_websocket_token = v4PrivatePostProfileWebsocketToken = Entry('profile/websocket_token', ['v4', 'private'], 'POST', {})
    v4_private_post_convert_estimate = v4PrivatePostConvertEstimate = Entry('convert/estimate', ['v4', 'private'], 'POST', {})
    v4_private_post_convert_confirm = v4PrivatePostConvertConfirm = Entry('convert/confirm', ['v4', 'private'], 'POST', {})
    v4_private_post_convert_history = v4PrivatePostConvertHistory = Entry('convert/history', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_create = v4PrivatePostSubAccountCreate = Entry('sub-account/create', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_delete = v4PrivatePostSubAccountDelete = Entry('sub-account/delete', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_edit = v4PrivatePostSubAccountEdit = Entry('sub-account/edit', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_list = v4PrivatePostSubAccountList = Entry('sub-account/list', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_transfer = v4PrivatePostSubAccountTransfer = Entry('sub-account/transfer', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_block = v4PrivatePostSubAccountBlock = Entry('sub-account/block', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_unblock = v4PrivatePostSubAccountUnblock = Entry('sub-account/unblock', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_balances = v4PrivatePostSubAccountBalances = Entry('sub-account/balances', ['v4', 'private'], 'POST', {})
    v4_private_post_sub_account_transfer_history = v4PrivatePostSubAccountTransferHistory = Entry('sub-account/transfer/history', ['v4', 'private'], 'POST', {})
